/*
 *
 *   OneWirePSK.h
 *
 *   PSK Modulator class.
 *
 *   Copyright (C) 2016 by Matt Roberts, KK5JY,
 *   All rights reserved.
 *
 */

#ifndef __KK5JY_ONEWIREPSK_H
#define __KK5JY_ONEWIREPSK_H

namespace KK5JY {
	namespace DSP {
		//
		//  One-wire PSK transmitter
		//
		template<int TxPin, int WaveTableLength = 8, int ShapeTableLength = 8>
		class OneWirePSK {
			private:
				// compute the timer2 divisor for a specific carrier frequency
				static uint32_t ComputeDivisor(double f0) {
					return round(F_CPU / (WaveTableLength * f0)) - 1;
				}

				// compute the carrier generated by a specific timer2 divisor
				static double ComputeF0(uint32_t divisor) {
					return F_CPU / static_cast<double>(WaveTableLength * (divisor + 1));
				}

				// compute the sampling rate
				static double ComputeFs(uint32_t divisor) {
					return F_CPU / divisor;
				}

			private:
				// storage for carrier waveform
				uint8_t wave[WaveTableLength];

				// storage for raised-cosine shaping waveform
				double shape[ShapeTableLength];

				// constants
				const double fc;			// center frequency
				const double bitRate;		// PSK bit rate
				const uint16_t div;			// the divisor
				const uint16_t bitPeriod;	// bit period, measured in TX samples
				const uint16_t bitMillis;	// bit period, measured in milliseconds
				const uint16_t leadPeriod;	// lead-in/out period, measured in samples
				const uint16_t leadInc;		// lead-in/out increment, samples per sample
				const uint16_t phaseWin;	// counter value for phase transition, both directions

				// ISR vars
				double m;				// multiplier; used for wave shaping
				volatile uint8_t index;	// wave table pointer
				uint8_t shidx;			// shape table pointer
				int8_t shinc;			// current shape table increment value (either +1 or -1)
				uint8_t mctr;			// sample counter for phase-inversion shaping
				uint8_t sctr;			// shape counter; increments for each shape step, every leadInc samples

				// this is the 'invert' flag; set to true when a phase inversion is desired;
				//    will be picked up by ISR and used to trigger 180-degree shift in output
				volatile bool invert;

			public:
				//
				//  ctor
				//
				//  f0 = carrier frequency
				//  br = bit rate (e.g., 31.25, 62.5, ...)
				//
				OneWirePSK(double f0 = 1500.0, double br = 31.25);

			public:
				void Invert() { invert = true; }
				void Setup();
				void Start();
				void Stop();

			public:
				void InterruptHandler();

			public:
				uint16_t BitMillis() const { return bitMillis; }
		};

		//
		//  OneWirePSK::ctor
		//
		template<int TxPin, int WaveTableLength, int ShapeTableLength>
		inline OneWirePSK<TxPin, WaveTableLength, ShapeTableLength>::OneWirePSK(double f0, double br)
			: fc(f0),
			  bitRate(br),
			  div(ComputeDivisor(fc)),
			  bitPeriod(ComputeFs(ComputeDivisor(div)) / static_cast<double>(bitRate)),
			  bitMillis(round(1000.0 / bitRate)),
			  leadPeriod(round(bitPeriod / 2.0)),
			  leadInc(round(leadPeriod / static_cast<double>(ShapeTableLength))),
			  phaseWin(round(2.0 * leadPeriod / leadInc)),
			  m(1.0),
			  index(0),
			  shidx(0),
			  shinc(1),
			  mctr(0),
			  sctr(0),
			  invert(false) {

			// populate the waveform table with a full-period sinusoid
			for (uint16_t i = 0; i != WaveTableLength; i++) {
				wave[i] = 128 + (127 * sin((2.0 * M_PI * i / WaveTableLength)));

				#ifdef VERBOSE_DEBUG
				Serial.print("wave[");
				Serial.print(static_cast<uint16_t>(i));
				Serial.print("]: ");
				Serial.println(static_cast<int16_t>(wave[i]));
				#endif
			}

			// populate the shaping table with a half-period raised cosine
			for (uint16_t i = 0; i != ShapeTableLength; i++) {
				shape[i] = 0.5 + (0.5 * cos((M_PI * i / ShapeTableLength)));

				#ifdef VERBOSE_DEBUG
				Serial.print("shape[");
				Serial.print(static_cast<uint16_t>(i));
				Serial.print("]: ");
				Serial.println(shape[i]);
				#endif
			}
		}

		//
		//  OneWirePSK::Setup()
		//
		template<int TxPin, int WaveTableLength, int ShapeTableLength>
		inline void OneWirePSK<TxPin, WaveTableLength, ShapeTableLength>::Setup() {
			// disable interrupts
			cli();

			//
			//  Run Timer1 in CTC mode, with 16-bit divisor; this drives the ISR
			//
			TCCR1A = 0;				// clear Timer1 control registers
			TCCR1B = 0;
			TCNT1 = 0;				// set Timer1 counter to zero
			TCCR1B |= (1 << WGM12); // enable CTC mode; limit is OCR1A
			TCCR1B |= (1 << CS10);  // set prescaler to 1x
			TIMSK1 = 0;				// Output Compare Interrupt Enable A (disabled at first)
			OCR1A = div;			// set clock divisor A
			OCR1B = 0;				// set clock divisor B

			//
			//  Run Timer2 in Fast PWM mode; drives pins 3, 11
			//
			pinMode(TxPin, OUTPUT); // configure pin for OUTPUT
			TCCR2A = 0;				// clear Timer2 control registers
			TCCR2B = 0;
			TCCR2B |= (1 << CS20);	// Set prescaler to 1x
			if (TxPin == 11)
				TCCR2A |= (1 << COM2A1);// Clear OC2A on compare match (pin 11)
			if (TxPin == 3)
				TCCR2A |= (1 << COM2B1);// Clear OC2B on compare match (pin 3)
			TCCR2A |= (				// Use 8-bit fast PWM mode; limit is 0xFF
					(1 << WGM20) |
					(1 << WGM21)
			);

			// enable interrupts
			sei();
		}

		//
		//  OneWirePSK::Start()
		//
		template<int TxPin, int WaveTableLength, int ShapeTableLength>
		inline void OneWirePSK<TxPin, WaveTableLength, ShapeTableLength>::Start() {
			// if already running, give up
			if (TIMSK1)
				return;

			// start at zero degrees
			m=1.0;
			index=0;
			shidx=0;
			shinc=1;
			mctr=0;
			sctr=0;

			// enable interrupts for timer overflow
			TIMSK1 = (1 << OCIE1A);
		}

		//
		//  OneWirePSK::Stop()
		//
		template<int TxPin, int WaveTableLength, int ShapeTableLength>
		inline void OneWirePSK<TxPin, WaveTableLength, ShapeTableLength>::Stop() {
			uint8_t n = WaveTableLength;

			// wait for the wave to drop to zero
			while ((index != 0) && (--n != 0)) {
				// nop
			}

			// disable interrupts for timer overflow
			TIMSK1 = 0;
		}

		//
		//  OneWirePSK::InterruptHandler()
		//
		template<int TxPin, int WaveTableLength, int ShapeTableLength>
		inline void OneWirePSK<TxPin, WaveTableLength, ShapeTableLength>::InterruptHandler() {
			// Update the PWM output
			if (TxPin == 11)				// pin 11 is driven by OCR2A
				OCR2A = m * wave[index];	// scale wave table entry by 'm'
			if (TxPin == 3)					// pin 3 is driven by OCR2B
				OCR2B = m * wave[index];	// scale wave table entry by 'm'

			// select next sample
			if (++index >= WaveTableLength) {
				index = 0;
			}

			// if inversion requested
			if (invert) {
				invert = false;		// clear inversion flag
				mctr = phaseWin;	// load the wave shaping sample counter
				sctr = leadInc;		// load the wave shaping pointer counter
				shidx = 0;			// point wave shaping to first cosine point
				shinc = 1;			// start by incrementing through cosine
				m = shape[0];		// load the first amplitude value
			}

			// run shaping
			if (mctr) {
				// decrement the shape increment counter
				if (--sctr == 0) {
					// reset and zero-compare the shape increment counter
					sctr = leadInc;

					// decrement the shape sample counter
					--mctr;

					// step to the next shape entry
					if (shidx || shinc == 1) {
						shidx += shinc;		// increment or decrement the shape pointer
						m = shape[shidx];	// and read the current amplitude from there
					} else {
						m = 1.0;	// if at the end of the shape envelope, reset magnitude to 1.0
					}

					// when at the point of minimum amplitude...
					if (shidx == ShapeTableLength - 1) {
						// count the other direction
						shinc = -1; // down

						// and phase invert
						index += (WaveTableLength / 2);
						if (index >= WaveTableLength) {
							index -= WaveTableLength;
						}
					}
				}
			}
		}
	}
}

#endif
