/*
 *
 *
 *    PskModem.ino
 *
 *    A simple terminal/modem for BPSK-31.
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2016 by Matthew K. Roberts, KK5JY. All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *
 *
 */

// ================== BEGIN CONFIGURATION SETTINGS ==================
//
//  It is not recommended to change any of these settings; the defaults
//  will provide a modem with a center frequency of 1500Hz, for BPSK-31.
//  This configuration will use A0 for audio input, D3 for audio output,
//  and D4 for PTT control.  The D13 LED will blink once per second, to
//  show normal program execution.
//

//
// Set the DATA output pin; this will generate a square-wave PSK signal
//    that must be band-pass filtered to create a clean PSK audio tone.
//
//    Valid values are 3 or 11, due to Timer/PWM limitations.
//
#define PSK_DATA_OUT_PIN (3)

//
// Set the PTT output pin; this pin will be driven LOW to activate the
//    radio PTT; set to zero to disable this feature; it is safe to
//    leave this enabled even if no hardware is connected to this pin.
//
#define PSK_PTT_PIN (4)

//
// Define the carrier frequency (in Hz):
//
#define PSK_DATA_FREQ_CENTER (1500)

//
// Set the bit rate of the data; standard BPSK-31 has a bit rate of 31.25.
//
#define PSK_DATA_BIT_RATE (31.25)

// Heartbeat LED - blinks at 1Hz to show normal loop activity
//    13 == 'L' LED on UNO; set to zero (0) to disable
#define HEARTBEAT_LED (13)

//
//  Host Port - assign a specific hardware port to be the
//              interface to the host; by default, the USB
//              connection on most boards is 'Serial', which
//              is also the hardware TTL serial port on the
//              UNO board.  The hardware serial device on
//              the Leonardo is Serial1.
//
#define HostSerial Serial

// the 'baud rate' of the serial port
#define SerialRate (9600)

//  Set this to determine the sample rate; divisor should be set
//  to one of the following values, depending on the sample rate
//  desired.  The recommended divisor is 128, giving a 9.6kSps
//  sampling rate.  The 16MHz AVR in the Arduino UNO does not have
//  sufficient CPU speed to run at a faster rate.
//
//  Divisor   ->    Sample Rate   ->   Nyquist Frequency
//    128            9615.4 S/s            4.81 kHz
//     64           19230.8 S/s            9.62 kHz
//     32           38461.5 S/s           19.23 kHz
//
#define PSK_DATA_SAMPLE_RATE_DIVISOR (128)

//
//  Define the zero-volts value to be read from the A/D.  If your
//  hardware has a perfectly centered DC offset (using a high-
//  quality voltage divider, etc.), just leave this at 128.  If
//  you have a larger DC offset, you can change this *slightly*
//  to compensate.
//
#define PSK_DATA_ANALOG_DC_OFFSET (128)

// ================== END OF CONFIGURATION SETTINGS ==================

// local headers
#include "CircularBuffer.h"
#include "Elapsed.h"
#define HEARTBEAT_ACTIVE (200) // 200ms
#include "SplitCommand.h"		// utility methods for cmd handling
#include "Heartbeat.h"			// code to blink the D13 LED
#include "OneWirePSK.h"			// PSK modulator
#include "avraudio.h"			// AVR audio input code
#include "pskem.h"				// PSK demodulator

// the version string
#define VERSION_STRING ("PskModem 1.0 (beta-20161111a)")

// CR+LF
#define LOCAL_CRLF ("\r\n")

// DEBUG: enable this to measure ISR runtime for performance tuning
//#define TIME_ISR_WORK

// import some namespaces
using namespace KK5JY::Collections;

//
//  Constants
//

// Limits
const uint32_t MaxPttLeadIn = 1000L;	// msec
const uint32_t MaxPttLeadOut = 1000L;	// msec
const uint32_t InputStringLength = 16;
const uint32_t CommandStringLength = 16;

// control keys
const char ESC = 0x1B;

////////////////////////////////////////////////////////////////
//
//  Globals
//
////////////////////////////////////////////////////////////////

// TX buffer - each element is an encoded byte with start/stop bits added
CircularBuffer<uint16_t> TxBuffer(32);

// TX shift register
static uint16_t TxShift = 0;

// TX preamble counter
static uint8_t pskTxZeroes = 0;

// a string to hold incoming data
String inputString;

// and another for command data
String cmdText = "";

// command parsing
String command = "";
String argument = "";

// whether the string is complete
bool inputReady = false;

// I/O buffer
char ioBuffer[32];

// echo USB RX data back to host
bool echo = false;

// currently parsing a command
bool cmd = false;

// TX mode (otherwise RX)
bool tx = false;

// TUNE mode
bool tune = false;

// tracks PTT state
bool pttState = false;

// global timer
volatile uint32_t now = 0;

// PTT timing
uint32_t pttLeadIn = 200;	// msec
uint32_t pttLeadOut = 750;	// msec

// Max idle time
uint32_t maxIdle = 1000;	// msec

// the bit period (in usec)
uint16_t bitPeriod = 0;

// tracks transmit bit edges
uint32_t pskTxEdge = 0;
uint32_t pskTxLastData = 0;

// the PSK receiver
typedef KK5JY::DSP::PskEmbedReceiver<int16_t, int32_t, int16_t> Receiver;
Receiver receiver(
	PSK_DATA_FREQ_CENTER,
	GetFs(PSK_DATA_SAMPLE_RATE_DIVISOR),
	PSK_DATA_BIT_RATE
);

// the varicode decoder
typedef KK5JY::DSP::PskEmbedDecoder Decoder;
Decoder decoder;

// PSK modulator
typedef KK5JY::DSP::OneWirePSK<PSK_DATA_OUT_PIN> Modulator;
Modulator pskmod(PSK_DATA_FREQ_CENTER, PSK_DATA_BIT_RATE);

////////////////////////////////////////////////////////////////
//
//  Support Methods
//
////////////////////////////////////////////////////////////////

// stores the audio level for diagnostics
static volatile int8_t audio_max = 0;
static volatile int16_t receiver_out = 0;
#ifdef TIME_ISR_WORK
static volatile uint32_t start;
static volatile uint16_t delayTimer = 0;
#endif

//
//  A/D Interrupt Handler -- called when new A/D value ready
//
ISR(ADC_vect) {
	static volatile int8_t sample = 0;	// the A/D input var
	static volatile uint8_t alevel = 0;	// the signal input level
	static volatile uint16_t rx = 0;		// the recieve data pointer

	#ifdef TIME_ISR_WORK
	start = micros();
	#endif

	//
	//  NOTE: BE VERY FAST HERE; this handler is being called
	//        thousands of times per second, and doing too
	//        much work here will cause samples to be lost.
	//

	// read the current A/D value and remove the DC offset
	sample = ADCH - PSK_DATA_ANALOG_DC_OFFSET;

	// store max audio level
	alevel = abs(sample);
	if (alevel > audio_max)
		audio_max = alevel;

	// do the demod
	rx = receiver.run(sample);
	if (rx != Receiver::PSK_EOF) {
		receiver_out = rx;
	}

	#ifdef TIME_ISR_WORK
	if (++delayTimer == 0)
		Serial.println(micros() - start);
	#endif
}


//
//  SetPskState(bool)
//
static inline void SetPskState(bool bit) {
	// if bit == 0, toggle phase
	if (!bit) {
		pskmod.Invert();
	}

	// store bit edge
	pskTxEdge = now;
}


//
//  GetNextTxBit(...)
//
//  Get the value of the next bit to transmit;
//  manipulate tx buffer and shift register as needed
//
static inline bool GetNextTxBit(bool & bit) {
	// if zero injection pending, count off one of them
	if (pskTxZeroes) {
		--pskTxZeroes;			// decrement the zero counter
		bit = false;			// store a zero bit
		pskTxLastData = now;	// store the last valid data bit time
		return true;			// success
	}

	// if the register is empty...
	if (TxShift == 0) {
		// try to get another word
		if ((!TxBuffer.Remove(TxShift)) || (TxShift == 0)) {
			// if that didn't work, the buffer is empty, so return EOF
			return false;
		}

		// then shift off the leading zeroes
		while ((TxShift & 0x8000) == 0) {
			TxShift <<= 1;
		}

		// queue up a zero
		pskTxZeroes = 1;

		// and send a zero
		bit = false;

		// success
		return true;
	}

	// store the next bit in 'bit'
	bit = (TxShift & 0x8000) == 0x8000;

	// and shift the TX register
	TxShift <<= 1;

	// store last valid data bit time
	pskTxLastData = now;

	// success
	return true;
}


////////////////////////////////////////////////////////////////
//
//  MAIN PROGRAM
//
////////////////////////////////////////////////////////////////

//
//  setup() - initialize hardware
//
void setup() {
	// start the serial port
	HostSerial.begin(SerialRate);
	// and wait for it to initialize
	while (!HostSerial) { /* nop */ }

	// allocate strings
	cmdText.reserve(CommandStringLength);
	inputString.reserve(InputStringLength);
	inputString = strcpy_P(ioBuffer, PSTR("#VERSION;"));
	inputReady = true;

	// PTT pin setup
	#if PSK_PTT_PIN > 0
	pinMode(PSK_PTT_PIN, OUTPUT);
	digitalWrite(PSK_PTT_PIN, LOW);
	#endif
	pttState = false;

	// HB LED
	#if HEARTBEAT_LED > 0
	HeartbeatSetup();
	#endif

	// calculate bit period (1000ms / bit time in ms)
	bitPeriod = static_cast<uint16_t>(round(1000.0 / PSK_DATA_BIT_RATE));

	// setup the A/D input
	ConfigureAnalogSampling(PSK_DATA_SAMPLE_RATE_DIVISOR);

	// configure the transmit output pin
	pskmod.Setup();
}


//
//  loop() - main program loop
//
void loop() {
	static bool hash = false;

	// read the clock
	now = millis();

	// blink the HB LED
	HeartbeatLoop();

	// process PTT key-up delay
	if (!tx && !tune && pttState) {
		uint32_t elapsed = Elapsed(now, pskTxLastData);
		// if the channel just idling...
		if (elapsed < maxIdle) {
			// diddle once every bit time
			if (Elapsed(now, pskTxEdge) >= bitPeriod) {
				SetPskState(false); // send zero for idle
			}
		}

		//
		//  if past maxIdle, but before maxIdle + pttLeadOut, just leave idle tone
		//
		
		// if past the max idle + pttLeadOut...
		else if (elapsed >= (pttLeadOut + maxIdle)) {
			// disable tone output
			pskmod.Stop();

			// disable PTT
			#if PSK_PTT_PIN > 0
			digitalWrite(PSK_PTT_PIN, LOW);
			#endif
			pttState = false;

			// re-enable the receiver sampling
			SetSamplingStatus(true);
		}
	} // !tx && pttState
	
	// if still transmitting, service the TX queue
	if (tx && !tune) {
		// re-read the clock
		now = millis();

		// if the current element has been sent...
		if (Elapsed(now, pskTxEdge) >= bitPeriod) {
			bool bit;
			if (GetNextTxBit(bit)) {
				// generate the leading edge
				SetPskState(bit);
			} else {
				tx = false;			// queue up a drop from TX
				SetPskState(false); // send zero for idle
			}
		}
	} // if (tx)...

	// if user input ready
	if (inputReady || hash) {
		// for each char received from the USB...
		for (int i = 0; hash || (i != inputString.length()); i += hash ? 0 : 1) {
			char ch = hash ? '#' : inputString.charAt(i);
			
			// if in command mode...
			if (cmd) {
				if (ch == '#') {
					// if a '#' sent mid-command, start over
					cmdText = "";
				} else if (ch == ';') {
					// terminate and process the command string
					cmd = false;
					bool ok = false;
					if (echo) HostSerial.print(';');

					// parse the command and argument
					SplitCommand(cmdText, command, argument);
					if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("DOC")))) {
						//
						//  QUERY DOC
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += strcpy_P(ioBuffer, PSTR(":UOS,ECHO,TUNE,"));
							cmdText += strcpy_P(ioBuffer, PSTR("PTTOUT,VERSION,"));
							cmdText += strcpy_P(ioBuffer, PSTR("LEVELMAX,LEVEL,"));
							cmdText += strcpy_P(ioBuffer, PSTR("PTTIN,PTTOUT"));
						} else {
							//
							//  TODO: respond to individual doc requests
							//
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("LEVELMAX")))) {
						// return the highest volume level since command last run
						ok = true;
						cmdText += '=';
						cmdText += 128;
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("LEVEL")))) {
						// return the highest volume level since command last run
						ok = true;
						cmdText += '=';
						cmdText += audio_max;
						audio_max = 0;
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("ECHO")))) {
						//
						//  QUERY/SET ECHO
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += echo ? "1" : "0";
						} else {
							int newEcho = argument.toInt();
							echo = newEcho ? true : false;
							ok = true;
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("TUNE")))) {
						//
						//  SET TUNE
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += tune ? "1" : "0";
						} else {
							if (argument.equals(strcpy_P(ioBuffer, PSTR("0")))) {
								// mute the tone
								pskmod.Stop();

								#if PSK_PTT_PIN > 0
								// de-PTT the radio
								digitalWrite(PSK_PTT_PIN, LOW);
								#endif
								pttState = false;

								// re-enable the receiver sampling
								SetSamplingStatus(true);

								tune = false;
								ok = true;
							} else if (argument.equals(strcpy_P(ioBuffer, PSTR("1")))) {
								// disable the receiver sampling
								SetSamplingStatus(false);

								// engage PTT
								pttState = true;
								#if PSK_PTT_PIN > 0
								digitalWrite(PSK_PTT_PIN, HIGH);

								// do PTT lead-in (TODO: find a way to do this without delay(...)
								delay(pttLeadIn);
								#endif

								// enable transmitter tone
								pskmod.Start();

								tune = true;
								ok = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("PTTIN")))) {
						//
						//  QUERY/SET PTT LEAD-IN
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += pttLeadIn;
						} else {
							uint32_t newPTT = argument.toInt();
							if (newPTT >= 0 && newPTT <= MaxPttLeadIn) {
								pttLeadIn = newPTT;
								ok = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("PTTOUT")))) {
						//
						//  QUERY/SET PTT LEAD-OUT
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += pttLeadOut;
						} else {
							uint32_t newPTT = argument.toInt();
							if (newPTT >= 0 && newPTT <= MaxPttLeadOut) {
								pttLeadOut = newPTT;
								ok = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("VERSION")))) {
						//
						//  QUERY VERSION NUMBER
						//
						ok = true;
						cmdText += '=';
						cmdText += strcpy_P(ioBuffer, PSTR(VERSION_STRING));
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("HASH")))) {
						//
						//  SEND HASH
						//
						hash = true;
						ok = true;
					}
					
					//
					//  TODO: process other command contents
					//

					// respond with #OK or #ERR
					if (echo)
						HostSerial.print(strcpy_P(ioBuffer, PSTR(LOCAL_CRLF)));
					HostSerial.print(ok ? strcpy_P(ioBuffer, PSTR("#OK:")) : strcpy_P(ioBuffer, PSTR("#ERR:")));
					HostSerial.print(cmdText);
					HostSerial.print(';');
					if (echo)
						HostSerial.print(strcpy_P(ioBuffer, PSTR(LOCAL_CRLF)));
					cmdText = "";
				} else {
					if (isAlpha(ch) || isDigit(ch) || isPunct(ch) || (ch == ' ')) {
						// add the char to the command string
						if (cmdText.length() < CommandStringLength)
							cmdText += ch;
						
						// echo the char, if enabled
						if (echo) {
							HostSerial.print(ch);
						}
					}
				}
				continue;
			}
			
			// start command mode??
			if (!hash) {
				if ((!cmd) && (ch == '#')) {
					cmd = true;
					if (echo) HostSerial.print('#');
					continue;
				}
			}

			// hard-stop?
			if (ch == ESC) {
				// turn off the tone
				pskmod.Stop();

				// PTT
				#if PSK_PTT_PIN > 0
				digitalWrite(PSK_PTT_PIN, LOW);
				pttState = false;
				#endif
				TxBuffer.Clear();
				tune = false;
				tx = false;

				// re-enable the receiver sampling
				SetSamplingStatus(true);

				continue;
			}

			// echo the char, if enabled
			if (hash) {
				hash = false;
			} else if (echo) {
				HostSerial.print(ch);
			}

			//
			//  TX - drive the output pin and PTT
			//
			if (!tune) {
				uint16_t w = varicode[static_cast<uint8_t>(ch)]; // encode the char to TX
				bool eok = TxBuffer.Add(w);
				if (! eok) {
					// send overflow message to console
					HostSerial.print(strcpy_P(ioBuffer, PSTR("#OVF;")));
				} else if (! tx) {
					bool bit;

					if (!pttState) {
						// send sequence of zeroes to synchronize receiver
						pskTxZeroes = 24; // 24 = ~0.75s for 31bps

						// disable the receiver sampling
						SetSamplingStatus(false);

						// engage PTT
						pttState = true;
						#if PSK_PTT_PIN > 0
						digitalWrite(PSK_PTT_PIN, HIGH);

						// do PTT lead-in (TODO: find a way to do this without delay(...)
						delay(pttLeadIn);
						#endif

						// enable transmitter tone
						pskmod.Start();
					}

					// start sending
					tx = true;

					// store most recent bit edge
					pskTxEdge = now;
				}
			}
		}

done:
		// clear the string:
		inputString = "";
		inputReady = false;
	}

	//
	//  RX: run the decoder
	//
	if (receiver_out != Receiver::PSK_EOF) {
		#ifndef TIME_ISR_WORK
		static uint16_t w;
		static uint8_t ch;

		w = receiver_out;
		receiver_out = Receiver::PSK_EOF;

		#if VERBOSE_DEBUG
		// DEBUG
		Serial.println(w);
		#endif
		ch = decoder.run(w);
		if (ch != Decoder::PSK_EOF) {
			if (ch == '#') {
				Serial.print("#HASH;");
			} else {
				Serial.print(static_cast<char>(ch));
			}
		}
		#endif
	}

	//
	//  HOST SERIAL/USB - service the serial port RX buffer
	//
	while (HostSerial.available() && (inputString.length() < InputStringLength)) {
		// get the new byte:
		char inChar = (char)HostSerial.read();
		// add it to the inputString:
		inputString += inChar;
		inputReady = true;
	}
}


//
//  Timer1 ISR - this is the timer that generates the sinusoid outputs
//
ISR(TIMER1_COMPA_vect) {
	pskmod.InterruptHandler();
}

// EOF
