/*
 *
 *
 *    ctrthr.h
 *
 *    Counter threshold trigger.
 *
 *    Copyright (C) 2016 by Matt Roberts.
 *    License: GNU GPL3 (www.gnu.org)
 *
 *
 */

#ifndef __CTRTHR_H
#define __CTRTHR_H

#include <stdlib.h>
#include <stdint.h>

namespace KK5JY {
	namespace DSP {
		//
		//
		//   a counter-based threshold class
		//
		//
		template <typename sample_t = int16_t, typename accum_t = int32_t>
		class CounterThreshold {
			public:
				// highest absolute value of the counter
				const accum_t limit;
				// level at which state change is done
				const accum_t trigger;
			private:
				// the counter
				accum_t counter;
				// the output state
				bool state;

			public: // ctors/dtors
				CounterThreshold(accum_t n, accum_t trig = 0);
				~CounterThreshold() { /* nop */ };
			private: // disallowed ctors
				CounterThreshold();
				CounterThreshold(const CounterThreshold &);

			public: // public methods
				// run one sample through the counter
				bool run(sample_t in);
				// return the value
				accum_t value() const { return counter; }
		};

		template <typename sample_t, typename accum_t>
		inline CounterThreshold<sample_t, accum_t>::CounterThreshold(accum_t n, accum_t trig)
			: limit(n), trigger(trig == 0 ? limit / 2 : trig), counter(0), state(false) {
			// nop
		}

		template <typename sample_t, typename accum_t>
		inline bool CounterThreshold<sample_t, accum_t>::run(sample_t in) {
			// run the counter
			if (in > 0) {
				counter++;
				if (counter > limit) counter = limit;
			} else if (in < 0) {
				counter--;
				if (counter < -limit) counter = -limit;
			}

			// if state change required, do it
			if ((!state) && (counter >= trigger)) {
				state = true;
			}
			if (state && (counter <= -trigger)) {
				state = false;
			}

			// return the new state
			return state;
		}
	}
}
#endif
