/*
 *
 *
 *    mdelay.h
 *
 *    Delay line multiplier.
 *
 *    Copyright (C) 2015-2016 by Matt Roberts.
 *    License: GNU GPL3 (www.gnu.org)
 *
 *
 */

#ifndef __KK5JY_MDELAY_H
#define __KK5JY_MDELAY_H

#include "fixed.h"

namespace KK5JY {
	namespace DSP {
		template <typename sample_t = double, typename frac_t = double>
		class DelayLineProduct {
			public:
				const double N;
				const size_t dist;

			private:
				const frac_t frac;
				const frac_t frac_1m;
				const size_t length;

				// delay line
				sample_t * const m_Delay;

				// input pointer
				sample_t * m_Ptr;

				// initializer
				void init() {
					m_Ptr = m_Delay;
					for (size_t i = 0; i != length; ++i) {
						*m_Ptr++ = static_cast<sample_t>(0);
					}
					m_Ptr = m_Delay;
				}

			public:
				DelayLineProduct(double f0, size_t fs);
				DelayLineProduct(double n);
				~DelayLineProduct() { delete[] m_Delay; }

			public:
				sample_t run(const sample_t value);
		};


		template <typename sample_t, typename frac_t>
		inline DelayLineProduct<sample_t, frac_t>::DelayLineProduct(double f0, size_t fs)
			: N(fs / f0),
			  dist(static_cast<size_t>(N)),
			  frac(N - dist),
			  frac_1m(1.0 - frac),
			  length(dist + 1),
			  m_Delay(new sample_t[length]),
			  m_Ptr(m_Delay) {
			init();
		}


		template <typename sample_t, typename frac_t>
		inline DelayLineProduct<sample_t, frac_t>::DelayLineProduct(double n)
			: N(n),
			  dist(static_cast<size_t>(N)),
			  frac(N - dist),
			  frac_1m(1.0 - frac),
			  length(dist + 1),
			  m_Delay(new sample_t[length]),
			  m_Ptr(m_Delay) {
			init();
		}


		template <typename sample_t, typename frac_t>
		inline sample_t DelayLineProduct<sample_t, frac_t>::run(const sample_t sample) {
			// delay line: fetch the sample one wavelength ago ('N' sample delay)
			sample_t *q = m_Ptr + dist;
			if (q >= (m_Delay + length)) {
				q -= length;
			}
			sample_t q1 = *q;

			// interpolate if needed, to get closer to the exact value of 'N'
			if (frac != 0) {
				q = m_Ptr + dist + 1;
				if (q >= (m_Delay + length)) {
					q -= length;
				}
				sample_t q2 = *q;
				q1 = (q2 * frac) + (q1 * frac_1m);
			}

			// compute the product of input and one wavelength ago
			sample_t product = sample * q1;

			// move the delay line pointer backwards through the delay line, and store the current sample
			*m_Ptr = sample;
			if (m_Ptr == m_Delay) {
				m_Ptr = m_Delay + length - 1;
			} else {
				--m_Ptr;
			}

			// done
			return product;
		}
	}
}

#endif // __KK5JY_MDELAY_H
