/*
 *
 *
 *    pskclock.h
 *
 *    PSK clock restoration.
 *
 *    Copyright (C) 2016 by Matt Roberts.
 *    License: GNU GPL3 (www.gnu.org)
 *
 *
 */

#ifndef __KK5JY_PSKCLOCK_H
#define __KK5JY_PSKCLOCK_H

#include "bitclock.h"

namespace KK5JY {
	namespace DSP {
		//
		//  BPSK-specific bit counter
		//
		template <typename sample_t = uint32_t, typename exp_t = double>
		class PskBitClock {
			public:
				const size_t N;
				const exp_t BitOffset;

			private:
				// generic bit counter
				BitClock<sample_t, exp_t> bc;

				// local counter - this is essentially the input counter to the clock
				sample_t bitCounter;

				// counter limit - this forms the clock
				sample_t next;

				// the current period
				sample_t period;

				// previous value -- used for differential input
				bool lastValue;

				// current state
				bool state;

			public:
				PskBitClock(sample_t n, sample_t margin, exp_t alpha);

				bool run(bool value);
		};

		//
		//  ctor
		//
		template <typename sample_t, typename exp_t>
		inline PskBitClock<sample_t, exp_t>::PskBitClock(sample_t n, sample_t margin, exp_t alpha)
			: N(n),
			  BitOffset(3.0 / 4.0),
			  bc(n, margin, alpha) {
			bitCounter = 0;
			next = N * BitOffset;
			period = N;
			lastValue = false;
		}

		//
		//  run
		//
		template <typename sample_t, typename exp_t>
		inline bool PskBitClock<sample_t, exp_t>::run(bool value) {
			bool result = false;

			// increment the period counter
			++bitCounter;

			if (bitCounter == next) {
				// time the next bit sample, in case it's not a zero
				next += period;

				// tell the caller to use this bit
				result = true;
#ifdef VERBOSE_DEBUG
				cerr
					<< "t = " << totalSamples
					<< "; value = " << value
					<< "; next = " << next << endl;
#endif
			}

			// state-change detection
			if (value != lastValue) {
				// if this is a rising edge...
				if (value) {
					// run the bit clock
					period = bc.run(bitCounter);

					// Store the next expected sample time; this will be the middle of the
					//   first half of the next bit-time; since a rising edge is the middle
					//   of the current bit-time, the middle of the first half of the next
					//   bit time is 3/4 bit periods from now...
					next = (BitOffset * period);

					// reset the bit counter
					bitCounter = 0;
#ifdef VERBOSE_DEBUG
					cerr
						<< "t = " << totalSamples
						<< "; period = " << period
						<< "; next = " << next << endl;
#endif
				}

				// store the value for next state-change test
				lastValue = value;
			}

			return result;
		}
	}
}

#endif
