/*
 *
 *
 *    pskem.h
 *
 *    PSK embedded modem components.
 *
 *    Copyright (C) 2016 by Matt Roberts.
 *    License: GNU GPL3 (www.gnu.org)
 *
 *
 */

#ifndef __KK5JY_PSKEM
#define __KK5JY_PSKEM

#ifdef __linux__
#include <stdlib.h>
#include <stdint.h>
#endif

#include "favg.h"
#include "mdelay.h"
#include "envdet.h"
#include "pskclock.h"
#include "varicode2.h"
#include "fixed.h"
#ifdef PSK_USE_PRESELECTOR
#include "resiir.h"
#endif

namespace KK5JY {
	namespace DSP {
		//
		//
		//   PSK decoder
		//
		//
		class PskEmbedDecoder {
			public:
				const static uint8_t PSK_EOF = 0x00;

			public:
				PskEmbedDecoder(void) { /* nop */ }

				uint8_t run(uint16_t shift);
		};


		//
		//  decoder::run
		//
		inline uint8_t PskEmbedDecoder::run(uint16_t shift) {
			uint8_t result = PSK_EOF;
			uint8_t i = 0;
			do {
				if (varicode[i] == shift) {
					shift = 0;
					result = i;
					break;
				}
			} while (i++ != 255);

			return result;
		}


		//
		//
		//   PSK receiver
		//
		//
		template
			<
			typename sample_t = int16_t,		// the signal type
			typename accum_t = int32_t,			// the integrator accumulator type
			typename bitclock_t = uint16_t		// the clock counter type
			>
		class PskEmbedReceiver {
			public:
				const static uint16_t PSK_EOF = 0x0000;

			private:
#ifdef PSK_USE_PRESELECTOR
				// the input preselector
				ResonantFilter<sample_t, FixedPoint<sample_t> > preselector;
#endif
				// the delay line -- one wavelength
				DelayLineProduct<sample_t, FixedPoint<sample_t> > delay;

				// the delay line integrator
				FixedAverager<sample_t, accum_t> integrator;

				// the envelope detector
				EnvelopeDetector<sample_t> detector;

				// the clock recovery logic
				PskBitClock<bitclock_t, FixedPoint<int16_t> > clock;

				// shift register for incoming words
				uint16_t shift;

			public:
				//
				//  ctor:
				//     f0   = center frequency
				//     fs   = sampling frequency
				//     rate = bit rate (e.g., 31.25, 62.5, ...)
				//
				PskEmbedReceiver(double f0, size_t fs, double rate);
				uint16_t run(sample_t);
		};

		//
		//  receiver::ctor
		//
		template <typename sample_t, typename accum_t, typename bitclock_t>
		inline PskEmbedReceiver<sample_t, accum_t, bitclock_t>::PskEmbedReceiver(double f0, size_t fs, double rate) :
#ifdef PSK_USE_PRESELECTOR
			  preselector(f0 / fs, 0.85),
#endif
			  delay(f0, fs),			// one wavelength
			  integrator((fs / f0)),	// one bit
			  detector(128, 16),
			  clock(
				fs / rate,				// period
				0.01 * (fs / rate),		// margin = 1% of period
				0.1),					// alpha
			  shift(0) {
			// nop
		}

		//
		//  receiver::run
		//
		template <typename sample_t, typename accum_t, typename bitclock_t>
		inline uint16_t PskEmbedReceiver<sample_t, accum_t, bitclock_t>::run(sample_t s) {
			uint16_t result = PSK_EOF;

			// run the signal chain: preselector -> delay -> integrator -> detector -> (bool output)
			bool out = detector.run(integrator.run(delay.run(
#ifdef PSK_USE_PRESELECTOR
				preselector.filter
#endif
				(s))));

			// run clock
			if (clock.run(out)) {
				// at this point, 'out' is a data bit so shift it in
				shift <<= 1;
				shift |= out;

				// decode
				if ((shift & 0b11) == 0) {
					result = shift;
					shift = 0;
				}
			}

			return result;
		}
	}
}

#endif
